/*****************************************************************************
 *
 * libwsw, Copyright (C) 2007 by Alexander Tsvyashchenko
 * contact address: <ejabberd@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef WSW_API_H
#define WSW_API_H

#include "core.h"

//------------------------------------------------------------------------------
namespace WSW { namespace Drivers { namespace XMPP { class Driver; } } }
namespace WSW { namespace API_details { class Task; } }
namespace SOCI { class Session; }
//------------------------------------------------------------------------------
namespace WSW
  {
  // This is the main, high-level interface to the library.
  //
  // For most operations two modes are supported:
  // * Synchronous, where the API call does not return until the operation
  //   is done.
  // * Asynchronous, where API call initiates operation in background / idle
  //   loop and returns immediately.
  // 
  // The API supports two major type of operations:
  // * indexed queries for items retrieval, where indices specify the range
  //   of items the user wants to retrieve
  // * general queries such as uploading collections
  //
  // For indexed queries the range is specified as [first_index, last_index]
  // where both indices are included in the result.
  // Indices can be negative - in this case they mean offset from the last item
  // minus one, for example [0, -1] specifies all items.
  //
  // Indices can be reversed - in this case items are fetched and delivered to
  // the user in reversed order, this is the most useful to get "last N items".
  //
  // For example, [-1, -10] means "get last 10 collections, starting from the
  // last one", while [-1, 0] means "get all collections in reverse order".
  //
  // In retrieval calls items are delivered to the user as soon as they're
  // available, one by one - however, until the call returns (or, for async
  // versions, until the status callback is called) these data must be treated
  // as temporary ones, as the call may fail in the end and you will have only
  // partial result.
  class API
    {
    public:
      // By 'generic error' in comments below I mean the following cases:
      //  * Network error, such as loss of connection, timeout or whatever.
      //  * The server error response, which typically means that either
      //    feature is not implemented.
      // Most likely I will change error interface later to be able to pass
      // more detailed information about errors to clients.
      enum Status
        {
        OK,
        // Empty or no JID was specified for current operation, but it's
        // required for the operation.
        NO_JID_SPECIFIED,
        // Indices specified for index request doesn't make sense, like
        // specifying interval laying completely outside of collections range.
        INVALID_INDICES, 
        // Change in data on server side was detected when performing operation,
        // for example due to actions of some other of your clients that changed
        // collections, due to auto-archiving or to collection expiration by
        // the server.
        // You should discard all information received so far from current
        // command execution and repeat the query.
        // If cached mode is enabled, on subsequent query replication will be
        // enforced. Also, the data that were requested so far will be not
        // re-requested once more, but returned from local cache, unless these
        // collections are indicated as changed during replication.
        CHANGE_DURING_OPERATION,
        // Generic error was detected when performing replication.
        ERROR_REPLICATION,
        // Generic  error was detected when issuing empty RSM statement to receive
        // items number. Note that items number is requested from server
        // only when it's really needed - in other words, if indices interval
        // specified in API call is not possible to translate to XEP-136 query
        // without knowing exact number of collections, such as (-10, -20).
        ERROR_GET_ITEMS_NUMBER,
        // Generic error was detected when receiving items from server.
        ERROR_GET_ITEMS,
        // Generic error was detected when receiving items from server.
        ERROR_UPDATE_ITEM,
        // Server error was detected - typically this means that the answer was
        // received and recognized as non-error answer, but content is wrong -
        // so either server is not XEP-136 compliant, the message was corrupted
        // or there's an error in libwsw code.
        ERROR_SERVER
        };

      // Collection index in [0, CollsNumber) range plus collection info itself.
      typedef std::pair<long, WSW::Core::CollectionInfo> CollectionInfoIndexed;
      // Message index in [0, MsgsNumber) range plus message itself.
      typedef std::pair<long, WSW::Core::Message> MessageIndexed;

      // Signatures of callbacks you should specify in API calls.

      typedef boost::function<void (const CollectionInfoIndexed&)> ReceiveCollection;
      typedef boost::function<void (const WSW::Core::CollectionLink& link, const MessageIndexed&)> ReceiveMessage;
      typedef boost::function<void (Status)> ReceiveStatus;

      struct Options
        {
#ifdef USE_SOCI
        // Use local caching when performing queries, so that fetched results
        // are stored in local DB and later answered from there, if applicable.
        // Periodically (see cache_ttl below) or immediately when discovering
        // server change replication with server is performed to mark those
        // collections which were changed / removed, and re-fetch them if they're
        // requested by the user.
        //
        // This is the default mode of libwsw operations and is recommended
        // to use in all cases unless you're running diskless client from
        // read-only storage, have very limited disk space or anything similar.
        bool use_db_cache;
#endif
        // If true, XEP-136, v 0.14 compliant replcation mode is used, but
        // this mode may lead to de-synchronization between client and server.
        // If false, newer mode is used, which is proposed as extention to
        // XEP-136, which is currently supported by mod_archive_odbc for ejabberd
        // server.
        bool use_legacy_replication;
        // If true, libwsw tries to guess about the action taken when not
        // enough info is available, typically - based on start time of collection.
        // Examples include uploading collection in the range that is not covered
        // by local cache - such upload may mean either creaion of new collection
        // or modification of existing one, and it's impossible to distinguish
        // between these without asking the server if it has this collection already
        // or not. However, most often users will not upload new collections
        // with the past date, and those collections which have the date in the
        // future compared to the last synchronization time are most likely
        // new ones, so this can be utilized to guess the right action.
        bool use_replication_heuristics;
        // Specifies max time in seconds between replications. If user performs
        // API call and libwsw finds out that last replication was done more than
        // cache_ttl seconds before, it will run replication before answering
        // user's request. Please note that if your server does support "changed"
        // extention to XEP-59 or if you're running only one client most of the
        // time, you can set this value quite high to speed up things.
        unsigned long cache_ttl;
        // Specifies how many modified items can be received per server request.
        // Larger values can speed up things a bit as it means smaller number
        // of requests is needed to fetch all modifications, but may violate
        // server "karma" for XMPP stanza length.
        unsigned long rsm_modified_limit;
        // Specifies how many collections can be received per server request.
        // Larger values can speed up things a bit as it means smaller number
        // of requests is needed to fetch all needed collections, but may
        // violate server "karma" for XMPP stanza length and mean that results
        // to the client will be sent in large bulks, which may be not desirable
        // for interactive applications.
        unsigned long rsm_collections_limit;
        // Specifies how many messages can be received per server request.
        // See rsm_collections_limit for trade-offs discussion.
        unsigned long rsm_messages_limit;

        Options();
        };

      // Constructor for API. It assumes ownership of driver an session
      // objects, so it will delete those upon destruction.
      // While currently construction is cheap, so in principle
      // API can be constructed as many times as needed, this is
      // not recommended, as it can change in the future. Minimally
      // you must make sure that it exists for the whole time the query
      // is running - this is important for async queries, for sync ones
      // you have not too much choice anyway ;-)
      API(WSW::Drivers::XMPP::Driver* driver,
#ifdef USE_SOCI
          SOCI::Session* session,
#endif
          const Options& opts);

      ~API();

      // Retrieves collections in the range [first_index, last_index] for
      // the specified JID by calling func() callback for each collection.
      // The JID must be non-empty. Returns status of the whole operation.
      Status getCollectionsSync(const std::string& jid,
                                long first_index,
                                long last_index,
                                const ReceiveCollection& func);

      // Retrieves collections in the range [first_index, last_index] for
      // the specified JID by calling func_coll() callback for each collection.
      // The JID must be non-empty. Upon operation completion calls func_status().
      void getCollectionsAsync(const std::string& jid,
                               long first_index, long last_index,
                               const ReceiveCollection& func_coll,
                               const ReceiveStatus& func_status);

      // Retrieves messages in the range [first_index, last_index] for
      // the specified collection by calling func() callback for each message.
      // Returns status of the whole operation.
      Status getMessagesSync(const WSW::Core::CollectionLink& link,
                             long first_index,
                             long last_index,
                             const ReceiveMessage& func_msg);

      // Retrieves messages in the range [first_index, last_index] for
      // the specified collection by calling func_msg() callback for each message.
      // Upon operation completion calls func_status().
      void getMessagesAsync(const WSW::Core::CollectionLink& link,
                            long first_index,
                            long last_index,
                            const ReceiveMessage& func_msg,
                            const ReceiveStatus& func_status);

      // Uploads collection to the server. Upload is done in single request,
      // so make sure you do not violate server limits on XMPP stanza size.
      // Returns status of the whole operation.
      // TODO: perform multiple requests??? Does anyone needs it???
      Status updateCollectionSync(const WSW::Core::Collection& collection);

      // Uploads collection to the server. Upload is done in single request,
      // so make sure you do not violate server limits on XMPP stanza size.
      // Upon operation completion calls func().
      void updateCollectionAsync(const WSW::Core::Collection& collection,
                                 const ReceiveStatus& func);

    private:
      WSW::Drivers::XMPP::Driver* driver_;
#ifdef USE_SOCI
      SOCI::Session* session_;
#endif
      Options opts_;

    private:
      Status executeSync_(WSW::API_details::Task* task);
    };
  };
//------------------------------------------------------------------------------

#endif // WSW_API_H
