/*****************************************************************************
 *
 * libwsw, Copyright (C) 2007 by Alexander Tsvyashchenko
 * contact address: <ejabberd@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "stdafx.h"
#include "api.h"
#include "api_private.h"
#include "task.h"

#ifdef USE_SOCI
#include "task_cached.h"
#endif

#include "utils.h"
#include "drivers/xmpp/driver.h"
//------------------------------------------------------------------------------
using namespace WSW::Core;
using namespace WSW::API_details;
using namespace WSW::Drivers::XMPP;
using namespace WSW::DateTime;
//------------------------------------------------------------------------------
// Update support
//------------------------------------------------------------------------------
namespace
  {
  class UpdateCollectionTaskBase: public virtual Task
    {
    public:
      UpdateCollectionTaskBase(Driver& driver, const Options& opts,
                               const Collection& collection,
                               const ReceiveStatus& func_status = ReceiveStatus()):
        Task(driver, opts, func_status),
        collection_(collection)
        {
        }
//----//------------------------------------------------------------------------
      void receiveStatus(Status status)
        {
        if (status == OK)
          finished();
        else
          cancel(WSW::API::ERROR_UPDATE_ITEM);
        }
//----//------------------------------------------------------------------------
    protected:
      Collection collection_;
    };
//--//--------------------------------------------------------------------------
  class UpdateCollectionDirectTask: public UpdateCollectionTaskBase
    {
    public:
      UpdateCollectionDirectTask(Driver& driver, const Options& opts,
                                 const Collection& collection,
                                 const ReceiveStatus& func_status = ReceiveStatus()):
        Task(driver, opts, func_status),
        UpdateCollectionTaskBase(driver, opts, collection, func_status) {}
//--//--------------------------------------------------------------------------
        virtual void init()
          {
          UpdateCollectionTaskBase::init();
          driver_.updateCollection(collection_, boost::bind(&UpdateCollectionTaskBase::receiveStatus, this, _1));
          }
//--//--------------------------------------------------------------------------
    };
//--//--------------------------------------------------------------------------
#ifdef USE_SOCI
//--//--------------------------------------------------------------------------
using namespace SOCI;
//--//--------------------------------------------------------------------------
  class UpdateCollectionCachedTask: public UpdateCollectionTaskBase, public TaskCached
    {
    public:
      UpdateCollectionCachedTask(Driver& driver, Session* session, const WSW::API::Options& opts,
                                 const Collection& collection,
                                 const ReceiveStatus& func_status = ReceiveStatus()):
        Task(driver, opts, func_status),
        TaskCached(driver, opts, func_status),
        UpdateCollectionTaskBase(driver, opts, collection, func_status)
        {
        setSession(session);
        }
//----//------------------------------------------------------------------------
      virtual void init()
        {
        UpdateCollectionTaskBase::init();
        upload();
        }
//----//------------------------------------------------------------------------
      std::string UpdateCollectionCachedTask::getLockingJID()
        {
        return collection_.info.self.with;
        }
//----//------------------------------------------------------------------------
      void performGetItemsNumberCacheQuery()
        {
        *session_ << "SELECT colls_number FROM users WHERE with_jid = :jid",
                     into(items_number_, ind_items_number_), use(jid_, "jid");
        }
//----//------------------------------------------------------------------------
      virtual void upload()
        {
        if (!startTransaction(boost::bind(&UpdateCollectionCachedTask::upload, this)))
          return;

        const CollectionLink& link = collection_.info.self;

        eIndicator ind;
        // Is this collection in cache?
        unsigned long coll_id = getCollectionIdByLink(ind, link);
        bool fl_update_indices = false;

        CollectionInfoCached info(collection_.info);
        unResolveLinks(info);
        info.dirty = false;

        if (ind != eOK)
          {
          fl_update_indices = cacheIsCompleteForLink(link.with, link.start);
          createCollectionByLink(link, fl_update_indices);
          coll_id = getCollectionIdByLink(ind, link);

          // We cannot be sure that this is the full collection - therefore,
          // we should mark it as "dirty", so that it's re-fetched if it
          // is requested later, and reset all subsequent collection's indices.
          //
          // We can use heuristics, though, to guess if this collection is new
          // or not.
          if (!fl_update_indices)
            {
            if (opts_.use_replication_heuristics)
              {
              // The collection is old - assume it exists already
              if (link.start <= last_update_)
                info.dirty = true;
              else // This is the new collection
                fl_update_indices = true;
              }
            else
              {
              info.dirty = true;
              resetCollectionsIndices(link);
              }
            }
          }
        else
          info.dirty = false;

        updateCollectionInCache(info, coll_id);

        if (fl_update_indices)
          updateIndicesAfter(link, +1);

        if (!info.dirty)
          {
          unsigned long num = 0;
          for (Collection::Messages::const_iterator it = collection_.messages.begin();
               it != collection_.messages.end(); ++it, ++num)
            {
            MessageCached info(*it);
            info.rsm_index = num;
            updateMessageInCache(info, coll_id);
            }
          }
        
        driver_.updateCollection(collection_, boost::bind(&UpdateCollectionCachedTask::receiveStatus, this, _1));
        }
//----//------------------------------------------------------------------------
    };
//--//--------------------------------------------------------------------------
#endif // USE_SOCI
//--//--------------------------------------------------------------------------
  }
//------------------------------------------------------------------------------
StatusAPI WSW::API::updateCollectionSync(const Collection& collection)
  {
  Task* task =
#ifdef USE_SOCI
    opts_.use_db_cache?
      static_cast<Task*>(new UpdateCollectionCachedTask(*driver_, session_, opts_, collection)):
#endif
      static_cast<Task*>(new UpdateCollectionDirectTask(*driver_, opts_, collection));

  return executeSync_(task);
  }
//------------------------------------------------------------------------------
void WSW::API::updateCollectionAsync(const Collection& collection, const ReceiveStatus& func_status)
  {
  Task* task =
#ifdef USE_SOCI
    opts_.use_db_cache?
      static_cast<Task*>(new UpdateCollectionCachedTask(*driver_, session_, opts_, collection, func_status)):
#endif
      static_cast<Task*>(new UpdateCollectionDirectTask(*driver_, opts_, collection, func_status));

  task->init();
  }
//------------------------------------------------------------------------------
