/*****************************************************************************
 *
 * libwsw, Copyright (C) 2007 by Alexander Tsvyashchenko
 * contact address: <ejabberd@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "stdafx.h"

#ifdef USE_SOCI

#include "api.h"
#include "api_private.h"
#include "task_cached.h"
#include "utils.h"
#include "drivers/xmpp/driver.h"

#define WSW_USE_MT_LOCKS

#ifdef WSW_USE_MT_LOCKS
#include <boost/thread/mutex.hpp>
#endif
//------------------------------------------------------------------------------
using namespace WSW::Core;
using namespace WSW::API_details;
using namespace WSW::Drivers::XMPP;
using namespace WSW::DateTime;
using namespace SOCI;
//----------------------------------------------------------------------------------
#ifdef WSW_USE_MT_LOCKS
typedef boost::mutex::scoped_lock scoped_lock;
#endif
//----------------------------------------------------------------------------------
namespace
  {
  std::map<std::string, TaskCached*> jids_owners_;
  std::map<std::string, std::list<ResumeFunction> > jids_funcs_;
  std::list<ResumeFunction> jids_funcs_by_modification_cmd_;

#ifdef WSW_USE_MT_LOCKS
  boost::mutex jids_access_;
#endif

  std::list<ResumeFunction> transaction_funcs_;
  TaskCached* transaction_owner_ = NULL;

#ifdef WSW_USE_MT_LOCKS
  boost::mutex transaction_access_;
#endif

  std::list<ResumeFunction> modification_cmd_funcs_;
  TaskCached* modification_cmd_owner_ = NULL;

#ifdef WSW_USE_MT_LOCKS
  boost::mutex modification_cmd_access_;
#endif
  }
//------------------------------------------------------------------------------
bool TaskCached::tryLockJID(const std::string& jid, const ResumeFunction& func)
  {
#ifdef WSW_USE_MT_LOCKS
  scoped_lock modification_cmd_funcs_access_lock(modification_cmd_access_);
  scoped_lock jids_access_lock(jids_access_);
#endif

  assert(!jid.empty());
  assert(jid == getLockingJID()); // currently we do not call it for other JIDs.
  assert(jids_owners_.find(jid) == jids_owners_.end() || jids_owners_[jid] != this);

  if (!modification_cmd_owner_)
    {
    if (jids_owners_.find(jid) == jids_owners_.end() || !jids_owners_[jid])
      {
      jids_owners_[jid] = this;
      return true;
      }

    jids_funcs_[jid].push_back(func);
    return false;
    }
  else
    {
    // Modification task is running - we will suspend the callee and put it into special queue
    // of tasks suspended due to modification command.
    jids_funcs_by_modification_cmd_.push_back(func);
    return false;
    }
  }
//------------------------------------------------------------------------------
void TaskCached::unlockJID(const std::string& jid)
  {
  ResumeFunction func;

    {
#ifdef WSW_USE_MT_LOCKS
    scoped_lock modification_cmd_access_lock(modification_cmd_access_);
    scoped_lock jids_access_lock(jids_access_);
#endif
   
    // Do we own lock for this JID at all?
    if (jids_owners_.find(jid) != jids_owners_.end() && jids_owners_[jid] == this)
      {
      jids_owners_.erase(jid);

      // Is there anyone else who waits for lock on this JID?
      if (jids_funcs_.find(jid) == jids_funcs_.end() || jids_funcs_[jid].empty())
        {
        if (!jids_funcs_by_modification_cmd_.empty())
          {
          func = jids_funcs_by_modification_cmd_.front();
          jids_funcs_by_modification_cmd_.pop_front();
          }
        }
      else
        {
        func = jids_funcs_[jid].front();
        jids_funcs_[jid].pop_front();
        }
      }
    }

  if (!func.empty())
    func();
  }
//------------------------------------------------------------------------------
bool TaskCached::tryLockForModificationCmd(const ResumeFunction& func)
  {
#ifdef WSW_USE_MT_LOCKS
  scoped_lock modification_cmd_funcs_access_lock(modification_cmd_access_);
  scoped_lock jids_access_lock(jids_access_);
#endif

  assert(!fl_replication_in_progress_);

  if (!modification_cmd_owner_ && jids_owners_.empty())
    {
    modification_cmd_owner_ = this;
    fl_replication_in_progress_ = true;
    return true;
    }

  modification_cmd_funcs_.push_back(func);
  return false;
  }
//------------------------------------------------------------------------------
void TaskCached::unlockForModificationCmd()
  {
  ResumeFunction func;

    {
#ifdef WSW_USE_MT_LOCKS
    scoped_lock modification_cmd_access_lock(modification_cmd_access_);
    scoped_lock jids_access_lock(jids_access_);
#endif

    assert (fl_replication_in_progress_ && modification_cmd_owner_ == this);

    modification_cmd_owner_ = NULL;
    fl_replication_in_progress_ = false;

    if (!modification_cmd_funcs_.empty())
      {
      func = modification_cmd_funcs_.front();
      modification_cmd_funcs_.pop_front();
      }
    else if (!jids_funcs_by_modification_cmd_.empty())
      {
      func = jids_funcs_by_modification_cmd_.front();
      jids_funcs_by_modification_cmd_.pop_front();
      }
    }

  if (!func.empty())
    func();
  }
//------------------------------------------------------------------------------
TaskCached::TaskCached(WSW::Drivers::XMPP::Driver& driver, const WSW::API::Options& opts,
                       const WSW::API::ReceiveStatus& func_status):
  Task(driver, opts, func_status),
  transaction_started_(false),
  modified_items_(0),
  modified_count_(0),
  fl_updated_(false),
  fl_replication_in_progress_(false)
  {
  }
//----------------------------------------------------------------------------------
bool TaskCached::startTransaction(const ResumeFunction& func)
  {
  assert(transaction_started_ || !func.empty());

  if (!transaction_started_)
    {
#ifdef WSW_USE_MT_LOCKS
    scoped_lock transaction_funcs_access_lock(transaction_access_);
#endif
    if (!transaction_owner_)
      {
      transaction_owner_ = this;
      session_->begin();
      transaction_started_ = true;
      return true;
      }
    else
      {
      transaction_funcs_.push_back(func);
      return false;
      }
    }
  return true;
  }
//----------------------------------------------------------------------------------
void TaskCached::transactionFinished()
  {
  ResumeFunction func;

    {
#ifdef WSW_USE_MT_LOCKS
    scoped_lock transaction_access_lock(transaction_access_);
#endif

    assert(transaction_started_ && transaction_owner_ == this);

    transaction_owner_ = NULL;

    if (!transaction_funcs_.empty())
      {
      func = transaction_funcs_.front();
      transaction_funcs_.pop_front();
      }
    }

  if (!func.empty())
    func();
  }
//----------------------------------------------------------------------------------
void TaskCached::cancel(StatusAPI res)
  {
  if (transaction_started_)
    {
    session_->rollback();

    if (res == WSW::API::CHANGE_DURING_OPERATION) // Force replication on next request.
      {
      session_->begin();
      forceReplicationOnNextRequest();
      session_->commit();
      }
    transactionFinished();
    }

  // Release modification command lock, if present
  if (fl_replication_in_progress_)
    unlockForModificationCmd();

  // In the case we had lock on our JID - release it
  unlockJID(getLockingJID());

  Task::cancel(res);
  }
//----------------------------------------------------------------------------------
void TaskCached::finished()
  {
  if (transaction_started_)
    {
    session_->commit();
    transactionFinished();
    }

  // In the case we had lock on our JID - release it
  unlockJID(getLockingJID());

  Task::finished();
  }
//----------------------------------------------------------------------------------
void TaskCached::getReplicationInfo()
  {
  replication_start_ = now();
  *session_ << "SELECT utc FROM last_update",
               into(last_update_, ind_last_update_);

  if (ind_last_update_ == eNull) // Forced replication
    {
    last_update_ = INF_NEG;
    fl_replication_needed_ = true;
    }
  else if (ind_last_update_ != eOK) // The value does not exist yet
    {
    last_update_ = replication_start_;
    fl_updated_ = true;
    fl_replication_needed_ = false;
    }
  else
    // If last update has expired already - perform replication to get the list of changes.
    fl_replication_needed_ = (replication_start_ - last_update_).total_seconds() > (long)opts_.cache_ttl;
  }
//----------------------------------------------------------------------------------
void TaskCached::forceReplicationOnNextRequest()
  {
  // Transaction should have been started already.
  startTransaction();

  *session_ << "UPDATE last_update SET utc = NULL";
  }
//----------------------------------------------------------------------------------
void TaskCached::init()
  {
  getReplicationInfo();

  if (fl_replication_needed_)
    {
    if (!tryLockForModificationCmd(boost::bind(&TaskCached::init, this)))
      return;

    // It's possible that in MT environment at current point modification command is no longer necessary,
    // thus check it once more.
    getReplicationInfo();

    if (fl_replication_needed_)
      {
      modified_start_ = last_update_;

      if (opts_.use_legacy_replication)
        modified_last_ = dateTime2Str(last_update_);

      setupReceiveModifications();
      }
    else
      unlockForModificationCmd();
    }
  else
    postReceiveModifications();
  }
//----------------------------------------------------------------------------------
void TaskCached::setupReceiveModifications()
  {
  Filter filter;

  RSM_Query rsm;
  rsm.limit = opts_.rsm_modified_limit;

  if (!opts_.use_legacy_replication)
    filter.start = modified_start_;

  if (!modified_last_.empty())
    {
    rsm.mode = RSM_Query::AFTER;
    rsm.value = modified_last_;
    }

  driver_.retrieveModifications(filter, rsm, boost::bind(&TaskCached::receiveModifications, this, _1, _2, _3));
  }
//----------------------------------------------------------------------------------
void TaskCached::updateIndicesAfter(const CollectionLink& in_link, long amount)
  {
  CollectionLink link(in_link); // Work-around for SOCI const bugs.

  // This function should be from receiveModifications only, which should establish transaction by this time
  startTransaction();

  unsigned long abs_amount = std::abs(amount);

  *session_ << "UPDATE collections SET rsm_index = rsm_index" << (amount > 0? '+' : '-') << " :amount " <<
               "WHERE with_jid = :with AND utc > :start AND rsm_index <> NULL",
               use(abs_amount, "amount"), use(link.with, "with"), use(link.start, "start");

  *session_ << "UPDATE users SET colls_number = colls_number" << (amount > 0? '+' : '-') << " :amount " <<
               "WHERE with_jid = :with",
               use(abs_amount, "amount"), use(link.with, "with");
  }
//----------------------------------------------------------------------------------
unsigned long TaskCached::getCollectionIdByLink(eIndicator& ind, const CollectionLink& in_link)
  {
  unsigned long id;
  CollectionLink link(in_link); // Work-around for SOCI const bugs.
  *session_ << "SELECT coll_id FROM collections WHERE with_jid = :with AND utc = :start",
               use(link.with), use(link.start), into(id, ind);
  return id;
  }
//----------------------------------------------------------------------------------
void TaskCached::createCollectionByLink(const CollectionLink& in_link, bool fl_calculate_index)
   {
   CollectionLink link(in_link); // Work-around for SOCI const bugs.
   int dirty = 1;

   if (fl_calculate_index)
     {
     // We need to determine the rsm index of new collection.
     unsigned long index;
     eIndicator ind_index;
     *session_ << "SELECT MAX(rsm_index) FROM collections WHERE with_jid = :with AND utc < :start",
                  use(link.with, "with"), use(link.start, "start"), into(index, ind_index);
        
     // Is this the first collection for this jid we know about?
     if (ind_index != eOK)
       index = 0;
     else
       index += 1;

     // We're called either from receiveModifications, from message's fetchCachedItems or from
     // child functions of setupReceiveItems. In all these cases transaction should be established already.
     startTransaction();

     // As we do not have full info about this collection - put only known fields and set "dirty" on.
     *session_ << "INSERT INTO collections(with_jid, utc, rsm_index, dirty) VALUES(:with, :start, :index, :dirty)",
                  use(link.with, "with"), use(link.start, "start"), use(index, "index"), use(dirty, "dirty");
     }
   else
     {
     startTransaction();

     // As we do not have full info about this collection - put only known fields and set "dirty" on.
     *session_ << "INSERT INTO collections(with_jid, utc, dirty) VALUES(:with, :start, :dirty)",
                  use(link.with, "with"), use(link.start, "start"), use(dirty, "dirty");
     }
   }
//----------------------------------------------------------------------------------
unsigned long TaskCached::unResolveLink(const CollectionLink& link)
  {
  eIndicator ind;
  unsigned long id = getCollectionIdByLink(ind, link);

  if (ind != eOK) // it's possible that this collection is not in cache yet - put it there then.
    {
    createCollectionByLink(link);
    id = getCollectionIdByLink(ind, link);
    }

  return id;
  }
//----------------------------------------------------------------------------------
void TaskCached::unResolveLinks(CollectionInfoCached& info)
  {
  if (info.coll.prev.isValid())
    info.prev_coll_id = unResolveLink(info.coll.prev);
  else
    info.prev_coll_id = CollectionInfoCached::INVALID_ID;

  if (info.coll.next.isValid())
    info.next_coll_id = unResolveLink(info.coll.next);
  else
    info.next_coll_id = CollectionInfoCached::INVALID_ID;
  }
//----------------------------------------------------------------------------------
void TaskCached::updateCollectionInCache(const CollectionInfoCached& in_info, unsigned long coll_id)
  {
  CollectionInfoCached info(in_info); // Work-around for SOCI const bugs.

  // As we're called after collection is fetched from server, transaction should have been established already
  // in setupReceiveItems.
  startTransaction();

  // We do not modify rsm_index here as it should not change, and when called from
  // messages task it is not filled.
  *session_ << "UPDATE collections SET with_jid = :with_jid, utc = :utc, " <<
               "prev_coll_id = :prev_coll_id, next_coll_id = :next_coll_id, " <<
               "subject = :subject, thread = :thread, crypt = :crypt, " <<
               "extra = :extra, dirty = :dirty " <<
               "WHERE coll_id = :coll_id",
               use(info), use(coll_id, "coll_id");
  }
//----------------------------------------------------------------------------------
void TaskCached::updateMessageInCache(const MessageCached& msg, unsigned long coll_id)
  {
  MessageCached info(msg); // Work-around for SOCI const bugs.
  // Is this collection already in our database?
  eIndicator ind;
  unsigned long msg_id;

  *session_ << "SELECT msg_id FROM messages WHERE coll_id = :coll_id AND rsm_index = :rsm_index",
               use(coll_id, "coll_id"), use(info.rsm_index, "rsm_index"), into(msg_id, ind);

  startTransaction();

  if (ind != eOK) // No such message found, insert it from scratch.
    *session_ << "INSERT INTO messages(coll_id, utc, dir, body, name, rsm_index) " <<
                 "VALUES(:coll_id, :utc, :dir, :body, :name, :rsm_index)",
                 use(coll_id, "coll_id"), use(info);
  else
    *session_ << "UPDATE messages SET coll_id = :coll_id, utc = :utc, " <<
                 "dir = :dir, body = :body, name = :name " <<
                 "WHERE id = :msg_id",
                 use(coll_id, "coll_id"), use(info), use(msg_id, "msg_id");
  }
//--//----------------------------------------------------------------------------
bool TaskCached::cacheIsCompleteForLink(const std::string& with, const TimeType& start)
  {
  CollectionLink link(with, start);

  unsigned long coll_ind_low;
  eIndicator ind_low;
  *session_ << "SELECT MAX(rsm_index) FROM collections WHERE with_jid = :with AND utc < :start",
               use(link.with, "with"), use(link.start, "start"), into(coll_ind_low, ind_low);

  unsigned long coll_ind_high;
  eIndicator ind_high;
  *session_ << "SELECT MIN(rsm_index) FROM collections WHERE with_jid = :with AND utc > :start",
               use(link.with, "with"), use(link.start, "start"), into(coll_ind_high, ind_high);

  // There's no collections for this JID.
  if (ind_low != eOK && ind_high != eOK)
    return false;

  // This will be the new first collection.
  if (ind_low != eOK && coll_ind_high == 0)
    return true;

  // This will be the new last collection.
  if (ind_high != eOK)
    {
    if (!fl_items_number_defined_)
      performGetItemsNumberCacheQuery();
    if (ind_items_number_ == eOK && coll_ind_low == items_number_ - 1)
      return true;
    }

  // All other cases with either of indices undefined are not acceptable
  if (ind_low != eOK || ind_high != eOK)
    return false;

  return coll_ind_high - coll_ind_low == 1;
  }
//----------------------------------------------------------------------------------
void TaskCached::resetCollectionsIndices(const CollectionLink& link_in)
  {
  CollectionLink link(link_in);

  *session_ << "UPDATE collections SET rsm_index = NULL WHERE with_jid = :with AND utc > :start",
               use(link.with, "with"), use(link.start, "start");
  }
//----------------------------------------------------------------------------------
void TaskCached::addNewCollectionByLink(const std::string& with, const TimeType& start)
  {
  CollectionLink link(with, start);

  // We can create new collection with calculated index only in the case there's no 'gap' at the point
  // of new collection link insertion, as otherwise we don't know what index new collection should have.
  bool fl_update_indices = cacheIsCompleteForLink(with, start);

  if (!fl_update_indices)
    {
    // We have no way to determine whether this collection was added or changed without querying the server -
    // as a consequence, all indices of collections after this one become undefined, so we will have
    // to put these indices in undefined state.

    // If heuristics are allowed, though, we can try to guess the action taken ...
    if (opts_.use_replication_heuristics)
      {
      // This collection was created after the last replication, so we'll assume it was added.
      if (start > last_update_)
        fl_update_indices = true;
      // Otherwise we assume that collection was just updated.
      }
    else
      // Reset all indices of collections after the given one.
      resetCollectionsIndices(link);
    }

  createCollectionByLink(link, fl_update_indices);

  if (fl_update_indices)
    updateIndicesAfter(link, +1); // Update all indices affected, if "proper" addition took place.
  }
//----------------------------------------------------------------------------------
void TaskCached::processChangedCollectionLink(const std::string& with, const TimeType& start)
  {
  eIndicator ind;

  // "changed" may mean either "added" or "updated". Determine which is the case.
  unsigned long coll_id = getCollectionIdByLink(ind, CollectionLink(with, start));

  if (ind != eOK)
    addNewCollectionByLink(with, start);
  else
    {
    // Just mark collection as dirty.
    int dirty = 1;
    *session_ << "UPDATE collections SET dirty = :dirty WHERE coll_id = :coll_id",
                 use(dirty, "dirty"), use(coll_id, "coll_id");
    }
  }
//----------------------------------------------------------------------------------
void TaskCached::removeCollectionByLink(const std::string& with, const TimeType& start)
  {
  CollectionLink link(with, start);
  eIndicator ind;
  unsigned long coll_id = getCollectionIdByLink(ind, link);

  // We will need to remove the collection as well as all messages, and adjust links.
  *session_ << "DELETE FROM collections WHERE coll_id = :coll_id",
               use(coll_id, "coll_id");
  *session_ << "DELETE FROM messages WHERE coll_id = :coll_id",
               use(coll_id, "coll_id");
  *session_ << "UPDATE collections SET prev_coll_id = NULL WHERE prev_coll_id = :coll_id",
               use(coll_id, "coll_id");
  *session_ << "UPDATE collections SET next_coll_id = NULL WHERE next_coll_id = :coll_id",
               use(coll_id, "coll_id");

  updateIndicesAfter(link, -1);
  }
//----------------------------------------------------------------------------------
void TaskCached::receiveModifications(const Driver::Modifications& modifs, const RSM_Result& rsm, Status status)
  {
  if (status == OK)
    {
    for (Driver::Modifications::const_iterator it = modifs.begin(); it != modifs.end(); ++it)
      {
      // Skip all modifications performed by ourselves; if legacy replication is allowed - we
      // do not use "by" value even if it is present, as other parts of library could have already
      // acted differently seeing use_legacy_replication flag.
      if (!opts_.use_legacy_replication && it->by == driver_.getJID())
        continue;

      std::string with_bare = WSW::XMPP::getBareJID(it->link.with);

      if (it->type == Modification::CHANGED)
        {
        if (!startTransaction(boost::bind(&TaskCached::setupReceiveModifications, this)))
          return;

        // Perform 'changed' processing for collections with full JID
        processChangedCollectionLink(it->link.with, it->link.start);

        // The same for bare JID
        if (with_bare != it->link.with)
          processChangedCollectionLink(with_bare, it->link.start);
        }
      else
        {
        assert(it->type == Modification::REMOVED);

        if (!startTransaction(boost::bind(&TaskCached::setupReceiveModifications, this)))
          return;

        removeCollectionByLink(it->link.with, it->link.start);

        if (with_bare != it->link.with)
          removeCollectionByLink(with_bare, it->link.start);
        }
      }

    modified_last_ = rsm.last;
    modified_items_ += (unsigned long)modifs.size();

    if (rsm.count < modified_items_ || (modified_count_ && modified_count_ != rsm.count))
      {
      cancel(WSW::API::CHANGE_DURING_OPERATION);
      return;
      }

    modified_count_ = rsm.count;

    // Did we enumerate all modifications?
    if (rsm.count == modified_items_ || modifs.size() < opts_.rsm_modified_limit)
      {
      last_update_ = replication_start_;
      fl_updated_ = true;
      unlockForModificationCmd();

      // Pass execution further on chain.
      postReceiveModifications();
      }
    else
      setupReceiveModifications();
    }
  else
    cancel(WSW::API::ERROR_REPLICATION);
  }
//----------------------------------------------------------------------------------
#endif // USE_SOCI
//----------------------------------------------------------------------------------
