/*****************************************************************************
 *
 * libwsw, Copyright (C) 2007 by Alexander Tsvyashchenko
 * contact address: <ejabberd@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "stdafx.h"

#ifdef USE_SOCI

#include "api.h"
#include "api_private.h"
#include "task_indexed_cached.h"
#include "utils.h"
#include "drivers/xmpp/driver.h"

//----------------------------------------------------------------------------------
using namespace WSW::Core;
using namespace WSW::API_details;
using namespace WSW::Drivers::XMPP;
using namespace WSW::DateTime;
using namespace SOCI;
//------------------------------------------------------------------------------
TaskIndexedCached::TaskIndexedCached(WSW::Drivers::XMPP::Driver& driver, const WSW::API::Options& opts,
                                     const WSW::API::ReceiveStatus& func_status, long first_index, long last_index):
  Task(driver, opts, func_status),
  TaskIndexed(driver, opts, func_status, first_index, last_index),
  TaskCached(driver, opts, func_status)
  {
  }
//----------------------------------------------------------------------------------
void TaskIndexedCached::preGetItemsNumber()
  {
  // Lock our own JID
  if (!tryLockJID(getLockingJID(), boost::bind(&TaskIndexedCached::preGetItemsNumber, this)))
    return;

  performGetItemsNumberCacheQuery();

  if (ind_items_number_ == eOK)
    fl_items_number_defined_ = true;

  setupGetItemsNumber();
  }
//----------------------------------------------------------------------------------
void TaskIndexedCached::postReceiveModifications()
  {
  preGetItemsNumber();
  }
//----------------------------------------------------------------------------------
bool TaskIndexedCached::verifyRSM(const WSW::Core::RSM_Result& rsm, unsigned long received_size)
  {
  if (!TaskIndexed::verifyRSM(rsm, received_size))
    return false;

  if (rsm.changed != TIME_INVALID && rsm.changed >= last_update_)
    {
    cancel(WSW::API::CHANGE_DURING_OPERATION);
    return false;
    }

  return true;
  }
//----------------------------------------------------------------------------------
long TaskIndexedCached::advanceDir(long& index)
  {
  if (fl_reversed_)
    return --index;
  else
    return ++index;
  }
//----------------------------------------------------------------------------------
void TaskIndexedCached::setupReceiveItems()
  {
  assert(fl_items_number_defined_ || !isCachePosValid());

  if (fl_items_number_defined_)
    {
    // First - output items already found in database and not in "dirty" state.
    for (; cur_item_index_ >= first_ && cur_item_index_ <= last_ &&
           isCachePosValid() && cachedItem().rsm_index == cur_item_index_ &&
           !isDirty(); advanceDir(cur_item_index_), advanceCachePos())
      outputItemFromCache();

    if (cur_item_index_ > last_ || cur_item_index_ < first_)
      {
      // Great! We've finished answering the query.
      finalize();
      return;
      }

    // We're going to perform queries to server and, thus, will write to the database.
    // As later it can be hard to handle transaction set up failure, we will do it now.
    // Note that it's safe to re-enter setupReceiveItems till this point as we will just skip the
    // first section up to this line on second call.
    if (!startTransaction(boost::bind(&TaskIndexedCached::setupReceiveItems, this)))
      return;

    // Now - setup the next request to server for those items we need to fetch.
    first_item_index_ = cur_item_index_;

    for (;cur_item_index_ >= first_ && cur_item_index_ <= last_; advanceDir(cur_item_index_))
      if (isCachePosValid() && cachedItem().rsm_index == cur_item_index_)
        {
        if (!isDirty())
          break; // We've found the item that we can output using cache - therefore, that's
                 // the end of range for current request.
        advanceCachePos();
        }

    items_received_ = 0;
    items_len_ = std::abs(cur_item_index_ - first_item_index_);
    fl_items_len_defined_ = true;
    if (fl_reversed_)
      first_item_index_ = cur_item_index_ + 1;
    item_last_.clear();
    }
  else
    {
    // See comments above
    if (!startTransaction(boost::bind(&TaskIndexedCached::setupReceiveItems, this)))
      return;

    // Make sure we're not going to proceed again with queries generation when entering
    // setupReceiveItems second time. Use for that dummy value, as cur_item_index_ is
    // not used anywhere besides setupReceiveItems.
    cur_item_index_ = -1;
    items_received_ = 0;
    }

  setupReceiveItemsQuery();
  }
//----------------------------------------------------------------------------------
void TaskIndexedCached::receiveItemsQueryFinished()
  {
  setupReceiveItems();
  }
//----------------------------------------------------------------------------------
void TaskIndexedCached::fetchItems()
  {
  if (!calculateIndices())
    return;

  fetchCachedItems();

  if (!fl_reversed_)
    cur_item_index_ = first_;
  else
    cur_item_index_ = last_;

  // Now construct requests for those indices ranges we're missing.
  setupReceiveItems();
  }
//----------------------------------------------------------------------------------
void TaskIndexedCached::finalize()
  {
  // Finally, if everything went OK and we also performed update, modify last_update
  if (fl_updated_)
    {
    if (!startTransaction(boost::bind(&TaskIndexedCached::finalize, this)))
      return;

    if (ind_last_update_ != eOK)
      *session_ << "INSERT INTO last_update(utc) VALUES(:utc)",
                   use(last_update_, "utc");
    else
      *session_ << "UPDATE last_update SET utc = :utc",
                   use(last_update_, "utc");
    }

  finished();
  }
//----------------------------------------------------------------------------------
#endif // USE_SOCI
//----------------------------------------------------------------------------------
